"""Report generating system using pandoc"""
import re
from cmdy import pandoc, wkhtmltopdf
from pathlib import Path

RESOURCE_DIR     = Path(__file__).resolve().parent / 'resources'
DEFAULT_FILTERS  = ['filetable', 'modal']
DEFAULT_TEMPLATE = 'bootstrap'

def _replaceAll(regex, callback, string):
	matches = re.finditer(regex, string)
	ret = ''
	start = 0
	for match in matches:
		ret += string[start:match.start(0)]
		start = match.end(0)
		ret += callback(match.group(), *match.groups())
	return ret + string[start:]

class ProcReport:

	def __init__(self, rptfile):
		self.rptfile   = Path(rptfile)
		self.source, self.appendix = self._analysis()

	def _analysis(self):
		"""Get the citations and appendix"""
		lines = self.rptfile.read_text().splitlines()
		# get the references first
		# references should be placed at the bottom
		source   = None
		appendix = None
		citations = {}
		for line in lines:
			line = line.rstrip('\n')
			if line.startswith('## Appendix'): # appendix has to be on level2 anyway
				appendix = []
			else:
				matched = re.match(r'\[(\d+)\]: (.+)', line)
				if matched:
					citations[matched.group(1)] = matched.group(2).strip()
				elif appendix is None:
					source = source or []
					source.append(line)
				else:
					appendix.append(line)

		# replace all citation marks with real references
		def replace(m):
			index = m[1:-1]
			if index not in citations:
				return m
			return '[#REF: %s #]' % citations[index]

		source = source and '\n'.join(source) or ''
		source = _replaceAll(r'\[\d+\]', replace, source)

		appendix = appendix and '\n'.join(appendix) or ''
		appendix = _replaceAll(r'\[\d+\]', replace, appendix)

		return source, appendix

class Report:

	def __init__(self, rptfiles, outfile, title):
		self.reports = [ProcReport(rptfile) for rptfile in rptfiles]
		self.srcpath = ':'.join(str(Path(rptfile).parent) for rptfile in rptfiles)
		self.outfile = Path(outfile)
		self.mdfile  = self.outfile.resolve().with_suffix('.md')
		if str(self.mdfile) in [str(Path(rptfile).resolve()) for rptfile in rptfiles]:
			# don't overwrite input file
			self.mdfile = self.outfile.with_suffix('.tmp.md')
		self.title   = title
		self.cleanup()

	def cleanup(self):
		citations = {}
		def replace(m):
			cite = m[7:-3]
			if cite not in citations:
				citations[cite] = len(citations) + 1
			return '<sup><a href="#REF_{i}">[{i}]</a></sup>'.format(i = citations[cite])

		with self.mdfile.open('w') as fmd:
			fmd.write('# %s\n\n' % self.title)
			appendix = ''
			for report in self.reports:
				# replace reference to citation indexes
				source = _replaceAll(r'\[#REF: .+? #\]', replace, report.source)
				if report.appendix:
					appendix += _replaceAll(r'\[#REF: .+? #\]', replace, report.appendix)

				fmd.write(source + '\n\n')

			if appendix:
				fmd.write('## Appendix\n')
				fmd.write(appendix + '\n\n')

			if citations:
				fmd.write('## Reference\n')
				for cite, index in sorted(citations.items(), key = lambda item: item[1]):
					fmd.write('<a name="REF_{i}" class="reference">**[{i}]** {cite}</a>\n\n'.format(
						i=index, cite=cite))

	def generate(self, standalone, template, filters):
		from pyppl import __version__ as pyppl_version
		from . import __version__ as report_version
		
		template = template or DEFAULT_TEMPLATE
		if template and '/' not in template:
			template = RESOURCE_DIR / 'templates' / template / 'template.html'

		ext   = self.outfile.suffix.lower()
		if 'htm' not in ext and 'pdf' not in ext:
			raise ValueError('Only html and pdf format supported currently.')

		args = (self.mdfile, )
		kwargs = {
			'metadata': [
				'pagetitle=%s' % self.title,
				'pyppl_version=%s' % pyppl_version,
				'report_version=%s' % report_version,
				'pdf=%s' % bool('pdf' in ext)],
			'read'    : 'markdown',
			'write'   : 'html5',
			'template': template,
			'filter'  : [	RESOURCE_DIR / 'filters' / (filt + '.py')
							for filt in DEFAULT_FILTERS 
							if filt != 'modal' or 'pdf' not in ext # pdf doesn't do modal
						] + (filters or []),
			'toc'           : True,
			'toc-depth'     : 3,
			'self-contained': standalone,
			'resource-path' : self.srcpath + ':' + str(Path(template).parent),
			'_raise'        : True,
			'_sep'          : 'auto',
			'_dupkey'       : True,
		}
		if 'pdf' not in ext:
			kwargs['output'] = self.outfile
			kwargs['_hold'] = True
			return pandoc(*args, **kwargs)
		else:
			kwargs['_pipe'] = True
			return pandoc(*args, **kwargs) | wkhtmltopdf('-', _ = self.outfile, _raise = True, _hold = True)



"""
Panflute filter to pop up a modal

Click [here](#detailinfo) for details.

```modal
id: detailinfo
title: Detailed information
closebtn: true
size: default # small, large, xlarge
---
content of the modal
```

into:

<div class="modal fade" id="exampleModalCenter" tabindex="-1" role="dialog" aria-labelledby="exampleModalCenterTitle" aria-hidden="true">
  <div class="modal-dialog modal-dialog-centered" role="document">
	<div class="modal-content">
	  <div class="modal-header">
		<h5 class="modal-title" id="exampleModalCenterTitle">Modal title</h5>
		<button type="button" class="close" data-dismiss="modal" aria-label="Close">
		  <span aria-hidden="true">&times;</span>
		</button>
	  </div>
	  <div class="modal-body">
		...
	  </div>
	  <div class="modal-footer">
		<button type="button" class="btn btn-secondary" data-dismiss="modal">Close</button>
	  </div>
	</div>
  </div>
</div>
"""

import panflute as pf

def fenced_action(options, data, element, doc):
	modalid  = options.get('id', 'modal1')
	title    = options.get('title')
	closebtn = options.get('closebtn', True)
	size     = options.get('size', 'default')
	size2class = {
		'default': None,
		'small'  : 'modal-sm',
		'sm'     : 'modal-sm',
		'large'  : 'modal-lg',
		'lg'     : 'modal-lg',
		'xlarge' : 'modal-xl',
		'xl'     : 'modal-xl',
	}

	components = []
	if title:
		modal_header1 = pf.Header(pf.Str(title), classes=['modal-title'], level=5, identifier=modalid + 'Title')
		modal_header2 = pf.Div(
			pf.Div(pf.Para(pf.Str('x')), attributes = {'aria-hidden': "true"}),
			classes = ['close', 'button'],
			attributes = {
				'type': 'button',
				'data-dismiss': 'modal',
				'aria-label': 'Close'
			})
		components.append(pf.Div(modal_header1, modal_header2, classes = ['modal-header']))
	components.append(pf.Div(*data, classes = ['modal-body']))
	if closebtn:
		components.append(pf.Div(
			pf.Div(pf.Para(pf.Str('Close')), classes = ['button', 'btn', 'btn-secondary'],
			attributes = {
				'type': 'button',
				'data-dismiss': 'modal',
			}),
			classes = ['modal-footer']
		))
	modal_content = pf.Div(*components, classes = ['modal-content'])
	mainclasses = ['modal-dialog', 'modal-dialog-centered', 'modal-dialog-scrollable']
	sizeclass = size2class.get(size)
	if sizeclass:
		mainclasses.append(sizeclass)
	model_dialog = pf.Div(modal_content, classes = mainclasses, attributes = {'role': 'document'})

	return pf.Div(model_dialog, classes = ['modal', 'fade'], identifier = modalid, attributes = {
		'tabindex'       : '-1',
		'role'           : 'dialog',
		'aria-labelledby': modalid + 'Title',
		'aria-hidden'    : "true"
	})

def action(elem, doc):
	"""
	Turn [link](modal#modal1) to:
	<a data-toggle="modal" data-target="#modal1">link</a>
	"""
	if isinstance(elem, pf.Link) and elem.url.startswith('modal'):
		modalid = elem.url.split('#')
		modalid = 'modal1' if len(modalid) == 1 else modalid[1]
		elem.attributes['data-toggle'] = 'modal'
		elem.attributes['data-target'] = '#' + modalid
	elif isinstance(elem, pf.Div) and 'modal' in elem.classes:
		return fenced_action(dict(id = elem.identifier, **elem.attributes), elem.content, elem, doc)

def main(doc=None):
	return pf.run_filter(action, doc=doc)

if __name__ == '__main__':
	main()


import sys
from time import time
from pathlib import Path
import yaml
from pyppl.plugin import hookimpl, postrun, addmethod
from pyppl.logger import logger, THEMES, LEVELS_ALWAYS
from pyppl.utils import fs, Box, formatSecs
from pyppl.exception import ProcAttributeError
from cmdy import CmdyReturnCodeException
from .report import Report

__version__ = "0.1.5"

@hookimpl
def setup(config):
	for colors in THEMES.values():
		colors['REPORT'] = colors['DONE']
	LEVELS_ALWAYS.add('REPORT')
	config['report']  = ''
	config['envs'].update(report = Box(
		level = 2,
		pre   = '',
		post  = '',
	))

@hookimpl
def procSetAttr(proc, name, value):
	if name == 'report' and value.startswith('file:'):
		scriptpath = Path(value[5:])
		if not scriptpath.is_absolute():
			from inspect import getframeinfo, stack

			# 0: .../pyppl_report/pyppl_report/__init__.py
			# 1: .../pluggy/callers.py
			# 2: .../pluggy/manager.py
			# 3: .../pluggy/manager.py
			# 4: .../pluggy/hooks.py
			# 5: .../PyPPL/pyppl/proc.py
			# 6: /file/define/the/report
			# if it fails in the future, check if the callstacks changed from pluggy
			caller = getframeinfo(stack()[6][0])
			scriptdir = Path(caller.filename).parent.resolve()
			scriptpath = scriptdir / scriptpath
		if not scriptpath.is_file():
			raise ProcAttributeError(
				'Report template file does not exist: %s' % scriptpath)
		proc.config[name] = "file:%s" % scriptpath

@hookimpl
def procGetAttr(proc, name):
	"""Pre-calculate the attribute"""
	if name == 'report':
		return Path(proc.workdir) / 'proc.report.md'

@hookimpl
def procPostRun(proc):
	"""Generate report for the process"""
	fs.remove(proc.report)
	if not proc.config.report:
		return

	logger.debug('Rendering report template ...')
	report = proc.config.report
	if report.startswith ('file:'):
		tplfile = Path(report[5:])
		## checked at __setattr__
		## and we are not able to change it on the run
		# if not fs.exists (tplfile):
		#	raise ProcAttributeError(tplfile, 'No such report template file')
		logger.debug("Using report template: %s", tplfile, proc = proc.id)
		report = tplfile.read_text()

	report  = proc.template(report, **proc.envs)
	rptdata = Box(jobs = [], **proc.procvars)
	for job in proc.jobs:
		jobdata  = job.data
		datafile = job.dir / 'output' / 'job.report.data.yaml'
		data = {}
		data.update(jobdata.job)
		if datafile.is_file():
			with datafile.open() as fdata:
				data.update(yaml.safe_load(fdata))
		rptdata.jobs.append(Box(i = jobdata.i, o = jobdata.o, **data))

	rptenvs  = Box(
		level = 2,
		pre   = '',
		post  = '',
		title = proc.desc)
	rptenvs.update(proc.envs.get('report', {}))
	rptdata.title = rptenvs.title
	reportmd = report.render(rptdata).splitlines()

	codeblock = False
	appendix  = False
	for i, line in enumerate(reportmd):
		if line.startswith('## Appendix') or appendix:
			appendix = True
			continue
		if line.startswith('#') and not codeblock:
			reportmd[i] = '#' * (rptenvs.level - 1) + line
		elif codeblock:
			if line.startswith('```') and len(line) - len(line.lstrip('`')) == codeblock:
				codeblock = False
		elif line.startswith('```'):
			codeblock = len(line) - len(line.lstrip('`'))

	proc.report.write_text(
		proc.template(rptenvs.pre, **proc.envs).render(rptdata) + '\n\n' +
		'\n'.join(reportmd) + '\n\n' +
		proc.template(rptenvs.post, **proc.envs).render(rptdata) + '\n'
	)

@postrun
def pyppl_report(ppl, outfile = None,
	title = 'A report generated by pipeline powered by PyPPL',
	standalone = True, template = False, filters = False):
	"""@API
	Generate report for the pipeline.
	Currently only HTML format supported.
	@params:
		outfile (file): The report file.
		title (str): The title of the report.
			- Default: 'A report generated by pipeline powered by PyPPL'
		standalone (bool): A standalone html file? Default: True
	@returns:
		(PyPPL): The pipeline object itppl.
	"""
	timer = time()
	outfile = outfile or (Path('.') / Path(sys.argv[0]).stem).with_suffix(
		'%s.report.html' % ('.' + str(ppl.counter) if ppl.counter else ''))
	logger.report('Generating report using pandoc ...')
	reports = [proc.report for proc in ppl.procs if proc.report.exists()]
	cmd = Report(reports, outfile, title).generate(standalone, template, filters)
	try:
		logger.debug('Running: ' + cmd.cmd)
		cmd.run()
		logger.report('Time elapsed: %s, report generated: %s' % (formatSecs(time() - timer), str(outfile)))
	except CmdyReturnCodeException as ex:
		logger.error(str(ex))
		sys.exit(1)

@hookimpl
def pypplInit(ppl):
	addmethod(ppl, 'report', pyppl_report)



"""
Panflute filter to parse CSV in fenced YAML code blocks

```table
caption    : An awesome table
file       : /path/to/csv/tsv/file
header     : true
width      : 1
total_width: .8
align      : default
rows       : 10
cols       : 0
csvargs    :
	dialect: unix
	delimiter: "\t"
datatable: # false to disable
	"page-length": 10
	"search": true
```
"""
import io
import json
import csv
import panflute as pf

def formatNumber(number):
	if len(number) < 10 or number.isdigit():
		return number
	try:
		number = float(number)
		if number > 0 and number < .001:
			return '%.3E' % number
		return '%.3f' % number
	except (ValueError, TypeError):
		return number

def fenced_action(options, data, element, doc):
	# We'll only run this for CodeBlock elements of class 'table'
	caption     = options.get('caption', 'Untitled Table')
	caption     = [pf.Str(caption)]
	filepath    = options.get('file')
	has_header  = options.get('header', True)
	width       = options.get('width', 1)
	total_width = float(options.get('total_width', .8))
	align       = options.get('align', 'default')
	nrows       = options.get('rows', 0)
	ncols       = options.get('cols', 0)
	csvargs     = options.get('csvargs', {})
	dtargs      = options.get('dtargs', {})

	csvargs['dialect']   = csvargs.get('dialect', "unix")
	csvargs['delimiter'] = csvargs.get('delimiter', "\t").encode().decode('unicode_escape')

	f = io.StringIO(data) if data else open(filepath)
	try:
		reader = csv.reader(f, **csvargs)
		body = []
		for i, row in enumerate(reader):
			if nrows and i > nrows + 1:
				continue
			cells = [pf.TableCell(pf.Plain(pf.Str(formatNumber(x))))
				for k, x in enumerate(row) if not ncols or k < ncols]
			body.append(pf.TableRow(*cells))
	finally:
		f.close()

	ncols = len(row)
	if has_header:
		header = body.pop(0)
		if body and len(body[0].content) == len(header.content) + 1:
			header.content.insert(0, pf.TableCell(pf.Plain(pf.Str(''))))
	elif nrows:
		body = body[:nrows]

	if not isinstance(width, list):
		width = [width]
	if len(width) == 1 and ncols:
		width = width * ncols
	sumwidth    = float(sum(width))
	total_width = min(total_width, 1)
	width       = [float(wid)*float(total_width)/sumwidth for wid in width]

	if not isinstance(align, list):
		align = [align]
	if len(align) == 1 and ncols:
		align = align * ncols
	align = ['Align' + al.capitalize() for al in align]

	table = pf.Table(*body,
		header=header, caption=caption, width=width, alignment = align)
	if dtargs is False:
		return pf.Div(table, classes = ['tablewrapper'])
	else:
		return pf.Div(table, classes = ['tablewrapper', 'datatablewrapper'], attributes = {
			'data-datatable': json.dumps(dtargs)
		})

def main(doc=None):
	return pf.run_filter(pf.yaml_filter, tag='table', function=fenced_action, doc=doc)

if __name__ == '__main__':
	main()


# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pyppl_report']

package_data = \
{'': ['*'],
 'pyppl_report': ['resources/filters/*',
                  'resources/templates/bootstrap/*',
                  'resources/templates/bootstrap/static/*']}

install_requires = \
['cmdy',
 'panflute>=1.0.0,<2.0.0',
 'pyparam',
 'pyppl>=2.0,<3.0',
 'pyyaml>=5.0,<6.0']

entry_points = \
{'console_scripts': ['pyppl_report = pyppl_report.console:main']}

setup_kwargs = {
    'name': 'pyppl-report',
    'version': '0.1.5.post0',
    'description': 'A report generating system for PyPPL',
    'long_description': '# pyppl_report\n\n[![Pypi][3]][4] [![Github][5]][6] [![PyPPL][7]][1] [![PythonVers][8]][4] [![docs][9]][2] [![Travis building][10]][11] [![Codacy][12]][13] [![Codacy coverage][14]][13]\n\nA report generating system for [PyPPL][1]\n\n## Installation\nRequires pandoc 2.7+ (and wkhtmltopdf 0.12.4+ when creating PDF reports)\n\n`pyppl_report` requires `pandoc/wkhtmltopdf` to be installed in `$PATH`\n\n```shell\npip install pyppl_report\n```\n\n## Usage\n### Specifiation of template\n\n````python\npPyClone.report = """\n## {{title}}\n\nPyClone[1] is a tool using Probabilistic model for inferring clonal population structure from deep NGS sequencing.\n\n![Similarity matrix]({{path.join(job.o.outdir, "plots/loci/similarity_matrix.svg")}})\n\n```table\ncaption: Clusters\nfile: "{{path.join(job.o.outdir, "tables/cluster.tsv")}}"\nrows: 10\n```\n\n[1]: Roth, Andrew, et al. "PyClone: statistical inference of clonal population structure in cancer." Nature methods 11.4 (2014): 396.\n"""\n\n# or use a template file\n\npPyClone.report = "file:/path/to/template.md"\n````\n\n### Generating report\n```python\nPyPPL().start(pPyClone).run().report(\'/path/to/report\', title = \'Clonality analysis using PyClone\')\n```\n\n![Snapshort](https://pyppl_report.readthedocs.io/en/latest/snapshot.png)\n\n### Extra data for rendering\nYou can generate a `YAML` file named `job.report.data.yaml` under `<job.outdir>` with extra data to render the report template. Beyond that, `proc` attributes and `args` can also be used.\n\nFor example:\n`job.report.data.yaml`:\n```yaml\ndescription: \'A awesome report for job 1\'\n```\nThen in your template, you can use it:\n```markdown\n## {{jobs[0].description}}\n```\n\n### Showing tables with csv/tsv file\n\n````markdown\n```table\ncaption    : An awesome table\nfile       : /path/to/csv/tsv/file\nheader     : true\nwidth      : 1   # width of each column\ntotal_width: .8  # total width of the table\nalign      : default # alignment of each column\nrows       : 10  # max rows to show\ncols       : 0   # max cols to show, default: 0 (show all)\ncsvargs    : # csvargs for `csv.read`\n\tdialect: unix\n\tdelimiter: "\\t"\n````\n\nYou may also specify `width` and `align` for individual columns:\n```yaml\nwidth:\n  - .1\n  - .2\n  - .1\n```\n\n### References\n\nWe use `[1]`, `[2]` ... to link to the references, so HTML links have to be in-place (in the format of `[text](link)` instead of `[text][link-index]`). All references from different processes will be re-ordered and combined.\n\n## Advanced usage\n[ReadTheDocs][2]\n\n\n[1]: https://github.com/pwwang/PyPPL\n[2]: https://pyppl_report.readthedocs.io/en/latest/\n[3]: https://img.shields.io/pypi/v/pyppl_report?style=flat-square\n[4]: https://pypi.org/project/pyppl_report/\n[5]: https://img.shields.io/github/tag/pwwang/pyppl_report?style=flat-square\n[6]: https://github.com/pwwang/pyppl_report\n[7]: https://img.shields.io/github/tag/pwwang/pyppl?label=PyPPL&style=flat-square\n[8]: https://img.shields.io/pypi/pyversions/pyppl_report?style=flat-square\n[9]: https://img.shields.io/readthedocs/pyppl_report.svg?style=flat-square\n[10]: https://img.shields.io/travis/pwwang/pyppl_report?style=flat-square\n[11]: https://travis-ci.org/pwwang/pyppl_report\n[12]: https://img.shields.io/codacy/grade/d11ed42fad49405abee3a8b695420709?style=flat-square\n[13]: https://app.codacy.com/project/pwwang/pyppl_report/dashboard\n[14]: https://img.shields.io/codacy/coverage/d11ed42fad49405abee3a8b695420709?style=flat-square\n\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'url': 'https://github.com/pwwang/pyppl_report',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import sys
from os import path
from pyparam import params
from cmdy import CmdyReturnCodeException
from pyppl import Box
from pyppl.logger import logger
from .report import Report

params['in'].required = True
params['in'].desc = 'The input file.'
params.i = params['in']
params.i.type = list
params.out.desc = 'The output file. Default: <in>.html'
params.out.callback = lambda opt, ps: opt.setValue(path.splitext(ps.i.value[0])[0] + '.html') if not opt.value else None
params.o = params.out
params.filter = []
params.filter.desc = 'The filters for pandoc'
params.title = 'Untitled document'
params.title.desc = 'The title of the document.'
params.template = 'bootstrap'
params.template.desc = 'The template to use. Either standard template name or full path to template file.'

def main():
	opts = params._parse(dict_wrapper = Box)
	cmd = Report(opts.i, opts.o, opts.title).generate(True, opts.template, opts.filter)
	try:
		logger.info('Running: ' + cmd.pipedcmd)
		cmd.run()
		logger.info('Report generated: ' + str(opts.o))
	except CmdyReturnCodeException as ex:
		logger.error(str(ex))
		sys.exit(1)
